//+------------------------------------------------------------------+
//|                                                    Bootstrap.mqh |
//|                        Copyright 2023, MetaQuotes Software Corp. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2023, MetaQuotes Software Corp."
#property link      "https://www.mql5.com"
#include<Math\Alglib\specialfunctions.mqh>
#include<Math\Stat\Math.mqh>
#include<UniformRandom.mqh>


//+------------------------------------------------------------------+
//|Function pointer                                                  |
//+------------------------------------------------------------------+
typedef double(*BootStrapFunction)(double &in[],int stop=-1);
//+------------------------------------------------------------------+
//|Boot strap types                                                  |
//+------------------------------------------------------------------+
enum ENUM_BOOSTRAP_TYPE
  {
   ENUM_BOOTSTRAP_PIVOT=0,
   ENUM_BOOTSTRAP_PERCENTILE,
   ENUM_BOOTSTRAP_BCA
  };

//+------------------------------------------------------------------+
//|class for implementing bootstrap                                  |
//+------------------------------------------------------------------+
class CBootstrap
  {
private:
   //--- private properties
   ENUM_BOOSTRAP_TYPE m_boot_type;                           //method of bootstrap to be applied
   uint               m_replications;                        //number of booststrap replications
   double             m_rep_cal[];                           //container for calculated replications
   double             m_data[];                              //helper container for intermediate calculations
   double             m_shuffled[];                          //shuffled collection of m_data
   CUniFrand         *m_unifrand;                            //random number generator
   bool               m_initialized;                         //flag
   BootStrapFunction  m_function;                            //function pointer
   //---  private helper methods
   void               shuffle(const uint start=0);
   bool               percentile_boot(double &in_out_conf[]);
   bool               pivot_boot(double &in_out_conf[]);
   bool               bca_boot(double &int_out_conf[]);
   //---- public methods
public:
   //--- constructor
                     CBootstrap(const ENUM_BOOSTRAP_TYPE boot_type,const uint nboot,const BootStrapFunction function,double &in_samples[]);
   //--- destructor
                    ~CBootstrap(void);
   //---
   bool              CalculateConfidenceIntervals(double &in_out_conf[]);
   bool              GetBootStrapStatistics(double &out[]) { if(m_initialized) return(ArrayCopy(out,m_rep_cal)==int(m_replications)); else return m_initialized; }
  };
//+------------------------------------------------------------------+
//|Constructor                                                       |
//+------------------------------------------------------------------+
CBootstrap::CBootstrap(const ENUM_BOOSTRAP_TYPE boot_type,const uint nboot,const BootStrapFunction function,double &in_samples[])
  {
//--- set the function pointer
   m_function=function;
//--- optimistic initilization of flag
   m_initialized=true;
//--- set method of boostrap to be applied
   m_boot_type=boot_type;
//--- set number of boostrap iterations
   m_replications=nboot;
//---make sure there are at least 5 boostraps
   if(m_replications<5)
      m_initialized=false;
//--- initilize random number generator
   m_unifrand=new CUniFrand();
   if(m_unifrand!=NULL)
      m_unifrand.SetSeed(MathRand());
   else
      m_initialized=false;
//--- copy samples to internal buffer
   if(ArrayCopy(m_data,in_samples)!=ArraySize(in_samples))
     {
      Print("Data Copy error ", GetLastError());
      m_initialized=false;
     }
//--- initialize shuffled buffer
   if(ArrayCopy(m_shuffled,in_samples)!=ArraySize(in_samples))
     {
      Print("Data Copy error ", GetLastError());
      m_initialized=false;
     }
//--- set memory for bootstrap calculations container
   if(ArrayResize(m_rep_cal,(int)m_replications)!=(int)m_replications)
     {
      Print("Memory allocation error ", GetLastError());
      m_initialized=false;
     }
//--- check function pointer
   if(m_function==NULL)
     {
      Print("Invalid function pointer");
      m_initialized=false;
     }
  }
//+------------------------------------------------------------------+
//|Destructor                                                        |
//+------------------------------------------------------------------+
CBootstrap::~CBootstrap(void)
  {
//--- clean up
   ArrayFree(m_shuffled);
//---
   ArrayFree(m_data);
//---
   ArrayFree(m_rep_cal);
//---
   delete m_unifrand;
  }
//+------------------------------------------------------------------+
//|Helper that shuffles elements of m_data array                     |
//+------------------------------------------------------------------+
void CBootstrap::shuffle(const uint start=0)
  {
//---
   int k=0;
   int size=ArraySize(m_data);
//---
   for(int i=int(start); i<size; i++)
     {
      k=(int)(m_unifrand.RandomDouble()*size);
      //---
      if(k>=size)
         k=size-1;
      //---
      m_shuffled[i]=m_data[k];
     }
//---
   return;
  }
//+------------------------------------------------------------------+
//|Do bootstraps and calculate confidence intervals                  |
//+------------------------------------------------------------------+
bool CBootstrap::percentile_boot(double &in_out_conf[])
  {
//--- local buffers
   double conf[];
//--- copy input parameters
   if(ArrayCopy(conf,in_out_conf)<ArraySize(in_out_conf))
     {
      Print(__LINE__,":",__FUNCTION__," error ",GetLastError());
      return false;
     }
//--- prepare input buffer to recieve output
   if(ArrayResize(in_out_conf,ArraySize(conf)*2)<ArraySize(conf)*2)
     {
      Print(__LINE__,":",__FUNCTION__," error ",GetLastError());
      return false;
     }
//--- reset internal buffer
   ZeroMemory(m_rep_cal);
//--- do bootstrap
   for(uint i = 0; i<m_replications; i++)
     {
      //--- shuffle
      shuffle();
      //--- calculate parameter
      m_rep_cal[i]=m_function(m_shuffled);
      //---
     }
//--- sort the buffers
   if(!ArraySort(m_rep_cal) || !ArraySort(conf))
     {
      Print("Sorting failure in ",__FUNCTION__," ",GetLastError());
      return false;
     }
//--- output the confidence bounds
   for(int i=0; i<ArraySize(conf); i++)
     {
      //---
      int k = (int)((1.0 - conf[i])*(m_replications+1))-1;
      //---
      if(k<0)
         k=0;
      //---
      in_out_conf[i*2]=m_rep_cal[k];
      //---
      in_out_conf[(i*2)+1]=m_rep_cal[m_replications-1-k];
     }
//---
   return true;
  }

//+------------------------------------------------------------------+
//|Do bootstraps and calculate confidence intervals                  |
//+------------------------------------------------------------------+
bool CBootstrap::pivot_boot(double &in_out_conf[])
  {
//--- internal buffers
   double conf[];
//--- copy input parameters
   if(ArrayCopy(conf,in_out_conf)<ArraySize(in_out_conf))
     {
      Print(__LINE__,":",__FUNCTION__," error ",GetLastError());
      return false;
     }
//--- prepare input buffer to recieve output
   if(ArrayResize(in_out_conf,ArraySize(conf)*2)<ArraySize(conf)*2)
     {
      Print(__LINE__,":",__FUNCTION__," error ",GetLastError());
      return false;
     }
//--- reset internal buffer
   ZeroMemory(m_rep_cal);
//--- do bootstrap
   for(uint i = 0; i<m_replications; i++)
     {
      //---
      shuffle();
      //---
      m_rep_cal[i]=m_function(m_shuffled);
      //---
     }
//--- sort the buffers
   if(!ArraySort(m_rep_cal) || !ArraySort(conf))
     {
      Print("Sorting failure in ",__FUNCTION__," ",GetLastError());
      return false;
     }
//--- calculate unshuffled parameter value
   double og = m_function(m_data);
//--- output confidence bounds
   for(int i=0; i<ArraySize(conf); i++)
     {
      //---
      int k = (int)((1.0-conf[i])*(m_replications+1))-1;
      //---
      if(k<0)
         k=0;
      //---
      in_out_conf[(i*2)+1]=2.0 * og - m_rep_cal[k];
      //---
      in_out_conf[i*2]=2.0 * og - m_rep_cal[m_replications-1-k];
      //---
     }
//---
   return true;
  }

//+------------------------------------------------------------------+
//|Do bootstraps and calculate confidence intervals                  |
//+------------------------------------------------------------------+
bool CBootstrap::bca_boot(double &in_out_conf[])
  {
//--- internal buffers
   double x[];
   double conf[];
//--- copy input parameters
   if(ArrayCopy(conf,in_out_conf)<ArraySize(in_out_conf))
     {
      Print(__LINE__,":",__FUNCTION__," error ",GetLastError());
      return false;
     }
//--- prepare input buffer to recieve output
   if(ArrayResize(in_out_conf,ArraySize(conf)*2)<ArraySize(conf)*2)
     {
      Print(__LINE__,":",__FUNCTION__," error ",GetLastError());
      return false;
     }
//--- copy data from class buffer
   if(ArrayResize(x,ArraySize(m_data))<ArraySize(m_data))
     {
      Print(__LINE__,":",__FUNCTION__," error ",GetLastError());
      return false;
     }
//--- reset buffers
   ZeroMemory(m_rep_cal);
   ZeroMemory(in_out_conf);
//--- calculate original parameter
   double theta=m_function(m_data);
   uint count=0;
//--- do bootstraps whilst enumerating
   for(uint i = 0; i<m_replications; i++)
     {
      //---
      shuffle();
      //---
      m_rep_cal[i]=m_function(m_shuffled);
      //---
      if(m_rep_cal[i]<theta)
         count++;
     }
//---
   if(count>=m_replications)
      count=m_replications-1;
   if(count<=0)
      count=1;
//---
   double inv_z=CNormalDistr::InvNormalDistribution(double(count)/double(m_replications));
//---
   double last=m_data[ArraySize(m_data)-1];
   double theta_dot=0.0;
   double temp=0;
   for(int i=0; i<ArraySize(m_data); i++)
     {
      temp=m_data[i];
      m_data[i]=last;

      x[i]=m_function(m_data,ArraySize(m_data)-1);

      theta_dot+=x[i];

      m_data[i]=temp;
     }
//---
   theta_dot/=double(ArraySize(m_data));
   double numer=0,denom=0,diff=0;
//---
   for(int i=0; i<ArraySize(m_data); i++)
     {
      diff=theta_dot - x[i];
      temp=diff*diff;
      denom+=temp;
      numer+=temp*diff;
     }
//---
   denom=MathSqrt(denom);
   denom=denom * denom * denom ;
   double accel=numer / (6.0 * denom + 1.e-60) ;
//---
   if(!ArraySort(m_rep_cal) || !ArraySort(conf))
     {
      Print("Sorting failure in ",__FUNCTION__," ",GetLastError());
      return false;
     }
//---
   double zlo,zhi,alo,ahi;
//---
   for(int i=0; i<ArraySize(conf); i++)
     {
      //---
      zlo=CNormalDistr::InvNormalDistribution(1-conf[i]);
      //---
      zhi=CNormalDistr::InvNormalDistribution(conf[i]);
      //---
      alo=CNormalDistr::NormalDistribution(inv_z+(inv_z+zlo)/(1.0-accel*(inv_z+zlo)));
      //---
      ahi=CNormalDistr::NormalDistribution(inv_z+(inv_z+zhi)/(1.0-accel*(inv_z+zhi)));
      //---
      int k = (int)(alo*(m_replications+1))-1;
      //---
      if(k<0)
         k=0;
      //---
      in_out_conf[i*2]=m_rep_cal[k];
      //---
      k = (int)((1-ahi)*(m_replications+1))-1;
      //---
      if(k<0)
         k=0;
      //---
      in_out_conf[(i*2)+1]=m_rep_cal[m_replications-1-k];
      //---
     }
//---
   return true;
  }
//+------------------------------------------------------------------+
//| public method for calculating confidence intervals               |
//+------------------------------------------------------------------+
bool CBootstrap::CalculateConfidenceIntervals(double &in_out_conf[])
  {
//--- safety check
   if(!m_initialized)
     {
      ZeroMemory(in_out_conf);
      return m_initialized;
     }
//--- check input parameter values
   if(ArraySize(in_out_conf)<=0 ||
      in_out_conf[ArrayMaximum(in_out_conf)]>=1 ||
      in_out_conf[ArrayMinimum(in_out_conf)]<=0)
     {
      Print("Invalid input values for function ",__FUNCTION__,"\n All values should be probabilities between 0 and 1");
      return false;
     }
//--- do bootstrap based on chosen method
   switch(m_boot_type)
     {
      case ENUM_BOOTSTRAP_PIVOT:
         return pivot_boot(in_out_conf);
      case ENUM_BOOTSTRAP_PERCENTILE:
         return percentile_boot(in_out_conf);
      case ENUM_BOOTSTRAP_BCA:
         return bca_boot(in_out_conf);
      default:
         return false;
     }
//---
  }
